package Mythic::SSHAPI::Client;

use strict;
use warnings;

use File::HomeDir;
use IO::Select;
use IPC::Open3;
use JSON;
use Log::Log4perl;
use POSIX qw(waitpid);
use Time::HiRes qw(usleep);
use URI::Encode qw(uri_encode);

use Mythic::SSHAPI::Response;

=head2 C<< $clnt = Mythic::SSHAPI::Client->new('root@remote', 'id_keyfile') >>

Instantiate a new C<Mythic::SSHAPI::Client> object. The first argument
specifies the C<user@host> to connect as, and the second names the private key
to use. This must be a file in C<$HOME/.ssh/>

=cut

sub new {
    my $class = shift;
    my ($target, $id, $opts) = @_;

    $opts //= {};

    my $self = {
        target => $target,
        id => $id,
        caller_id => $opts->{caller_id}
    };
    bless $self, $class;
    if (defined($self->{id}) && ! -f $self->key_file()) {
        die "Key file ".$self->key_file()." does not exist";
    }
    return $self;
}

sub logger {
    if(!Log::Log4perl->initialized()) {
        Log::Log4perl->init({
            "log4perl.logger"                 => "INFO, Screen",
            "log4perl.appender.Screen"        => "Log::Log4perl::Appender::Screen",
            "log4perl.appender.Screen.stderr" => 1,
            "log4perl.appender.Screen.mode"   => "append",
            "log4perl.appender.Screen.layout" => "Log::Log4perl::Layout::SimpleLayout",

        });
    }
    return Log::Log4perl->get_logger();
}


sub key_file {
    my $self = shift;
    return File::HomeDir->my_home."/.ssh/".$self->{id};
}


sub do {
    my $self = shift;
    my ($path, $data, $opts) = @_;
    $opts //= {};
    my $p;
    if (ref $path) {
        $p = '/'.join('/', map { uri_encode($_, { encode_reserved => 1 }) } @$path );
    }
    else {
        $p = $path;
    }
    $data = defined($data) ? encode_json($data) : '{}';
    return $self->do_tries($p, $data, $opts);
}

sub forward_key {
    my $self = shift;
    my $agent = `ssh-agent`;
    $agent =~ m!SSH_AUTH_SOCK=(/tmp/ssh[^; ]+);!;
    $ENV{SSH_AUTH_SOCK} = $1;
    if (my $pid = fork) {
        waitpid $pid, 0;
    }
    else {
        close STDERR;
        open STDERR, '>', '/dev/null';
        exec 'ssh-add', $self->key_file();
    }
}


=head2 C<do_tries>

We try the client call C<scalar @tries> times before giving up. C<$tries[$n]>
says how long to sleep before the C<$n>th try, in fractional seconds. The
actual value of $tries[0] is not used, it's assumed to be zero.

C<$timeouts[$n]> says what connect timeout to use on the C<$n>th try.

=cut

# Allow 5s the first time in case the host is a bit busy.  Subsequent retries
# are in case host is DoSed by SSH brute force, so reduce timeout in order to keep
# total time to giving up low.
my @tries = (0, 1, 1);
my @timeouts = (5, 2, 2);

sub do_tries {
    my $self = shift;
    my ($path, $data, $opts) = @_;
    $opts //= {};

    # We need this because $opts doesn't get reset
    # between tries so if we were to check whether
    # $opts->{timeout} is defined inside the loop,
    # we would end up only ever using the first
    # value in @timeouts.
    my $use_default_timeouts = !defined($opts->{timeout});

    my $res;
    for (my $n = 0; $n < @tries; ++$n) {
        if ($n > 0) {
            my $usecs = 1E6 * $tries[$n];
            usleep $usecs;
        }

        if ($use_default_timeouts) {
            $opts->{timeout} = $timeouts[$n];
        }

        $res = $self->do_raw($path, $data, $opts);
        if (!$res->success()) {
            logger()->error("Error executing API command on $self->{target}: ".$res->error());
            logger()->error("error output: ". $res->{stderr});
        }
	elsif($res->stderr()) {
            logger()->warn("Non empty stderr executing API command on $self->{target}");
            logger()->warn("error output: ". $res->{stderr});
	}
        if (!$res->success() && $res->exit_code() == 255) {
            logger()->error(sprintf("Retrying failed SSH API command on %s (%d of %d)",$self->{target},$n,scalar(@tries)));
        }
        else {
            last;
        }
    }
    return $res;
}

sub __build_command {
    my $self = shift;
    my ($path, $data, $opts) = @_;
    $opts //= {};

    my @command;
    if ($self->{id}) {
        push @command, "ssh";
        if ($opts->{forward}) {
            push @command, "-A";
        }
        push @command, ("-o", "IdentitiesOnly=yes", "-i", $self->key_file());
        push @command, ("-o", "SendEnv=SSHAPI_CALLER");
        push @command, ("-o", "VerifyHostKeyDNS=yes");
        if (defined($opts->{timeout})) {
            push @command, ("-o", "ConnectTimeout=" . $opts->{timeout});
        }
    }
    push @command, ($self->{target}, $path);

    return @command;
}

sub do_raw {
    my $self = shift;
    my ($path, $data, $opts) = @_;
    $opts //= {};

    logger()->info("Executing SSH API command on $self->{target}: $path");

    if ($self->{caller_id}) {
        $ENV{SSHAPI_CALLER} = $self->{caller_id};
    }

    if ($opts->{forward}) {
        $self->forward_key();
    }

    my @command = $self->__build_command($path, $data, $opts);
    my $pid = open3(*I, *O, *E, @command);

    my $wfh = new IO::Select();
    $wfh->add(\*I);

    my $rfh = new IO::Select();
    $rfh->add(\*O);
    $rfh->add(\*E);

    my $to_write = length($data);
    my $obuf = '';
    my $ebuf = '';

    while ($rfh->count() + $wfh->count() > 0) {
        my ($r, $w, undef) = IO::Select::select($rfh, $wfh, undef);
        if (defined($w) && @$w) {
            my $written = syswrite(I, $data);
            $to_write -= $written;
            if($to_write == 0 || !$written) {
                $wfh->remove(\*I);
                close I;
            }
        }
        if (defined($r) && @$r) {
            foreach my $fh (@$r) {
                if ($fh == \*O) {
                    my $read = sysread($fh, $obuf, 10000000, length($obuf));
                    if ($read == 0) {
                        $rfh->remove(\*O);
                    }
                }
                elsif ($fh == \*E) {
                    my $read = sysread($fh, my $tebuf, 10000000, 0);
                    if ($read == 0) {
                        $rfh->remove(\*E);
                    }
                    else {
                        while ($tebuf =~ s/^SSH-API-PROGRESS:([^\n]+)$//ms) {
                            if ($opts->{progress}) {
                                &{$opts->{progress}}($1);
                            }
                        }
                        $ebuf .= $tebuf;
                    }
                }
            }
        }
    }
    waitpid $pid, undef;
    return Mythic::SSHAPI::Response->new($obuf, $ebuf, $?/256);
}

sub do_streaming {
    my $self = shift;

    my ($path, $data, $opts) = @_;

    $data //= '';
    $opts //= {};

    logger()->info("Executing SSH API command on $self->{target}: $path");

    if ($self->{caller_id}) {
        $ENV{SSHAPI_CALLER} = $self->{caller_id};
    }

    if ($opts->{forward}) {
        $self->forward_key();
    }

    my @command = $self->__build_command($path, $data, $opts);

    # close STDIN, OUT and ERR
    #close(\*STDIN);
    #close(\*STDOUT);
    #close(\*STDERR);

    # Make new STDIN, OUT, ERR
    my $in = IO::Handle->new();
    my $out = IO::Handle->new();
    my $err = IO::Handle->new();

    my $pid = open3($in, $out, $err, @command);

    my $wfh = new IO::Select();
    $wfh->add($in);

    my $rfh = new IO::Select();
    $rfh->add($out);
    $rfh->add($err);

    my $to_write = length($data);
    my $obuf = '';
    my $ebuf = '';

    while ($to_write > 0) {
        my (undef, $w, undef) = IO::Select::select(undef, $wfh, undef);

        if (defined($w) && @$w) {
            my $written = syswrite(I, $data);
            $to_write -= $written;
        }
    }

    return ($pid, $rfh, $wfh, $out, $err, $in);
}

1;
