package Mythic::SSHAPI::Server;

use strict;
use JSON;
use Data::Dumper;
use Readonly;
use URI::Encode qw(uri_decode);

Readonly::Scalar our $NO_PAYLOAD => 0;
Readonly::Scalar our $JSON_PAYLOAD => 1;
Readonly::Scalar our $RAW_PAYLOAD => 2;

sub new {
    my $class = shift;
    my $self = {
        endpoints => [],
        filters => [],
    };
    bless $self, $class;
    $self->add_endpoint("/help", \&help);
    return $self;

}

sub help {
    my $self = shift;
    return ({ endpoints => [ sort map { $_->{path} } @{$self->{endpoints}} ] });
}

sub add_endpoint {
    my $self = shift;
    my ($path, $callback, $payload) = @_;
    push @{$self->{endpoints}}, {
        path => $path,
        callback => $callback,
        payload => ($payload || $NO_PAYLOAD),
    }
}

sub add_filter {
    my $self = shift;
    my ($path, $callback) = @_;
    push @{$self->{filters}}, {
        path => $path,
        callback => $callback
    }
}

sub path_match {
    my $self = shift;
    my($path, $test) = @_;
    my @parts;
    while ($path =~ s|/:([^/]+)|/([^/]*)|) {
        push @parts, $1;
    }

    if (my @matches = ($test =~ m/^$path$/)) {
        if (!@parts) {
            return {}
        }
        my $params = {};
        for (my $i = 0; $i < @matches; $i++) {  
            $params->{$parts[$i]} = uri_decode($matches[$i]);
        }
        return $params;
    }

    return undef;

}


sub go {
    my $self = shift;
    my $path = $ENV{SSH_ORIGINAL_COMMAND} || $ARGV[0] || ''; 
    my $endpoint;
    my $params;

    foreach my $filter (@{$self->{filters}}) {
        if($params = $self->path_match($filter->{path}, $path)) {
            my $filtered = &{$filter->{callback}}($params);
            if($filtered) {
                $self->do_response(undef, $filtered);
                return;
            }
        }
    }

    foreach my $ep (@{$self->{endpoints}}) {
        if($params = $self->path_match($ep->{path}, $path)) {
            $endpoint = $ep;
            last;
        }
    }

    if ($endpoint) {
        my $x;
        if ($endpoint->{payload} != $NO_PAYLOAD) {
            my $data = '';
            while (<STDIN>) {
                $data .= $_;
            }
            if($endpoint->{payload} == $JSON_PAYLOAD) {
              $x = decode_json($data);
            }
            else {
              $x = $data;
            }
        }

        $self->do_response(&{$endpoint->{callback}}($self, $params, $x));
        return;
    }
    else {
        $self->do_response(undef, "Endpoint '$path' not found", '404');
    }
}

sub do_response {
    my $self = shift;
    my ($res, $err, $errcode) = @_;

    my $o = {
        content => $res,
    };
    if ($err || $errcode ) {
        $o->{error} = { message => $err, code => $errcode };
    }
    my $j = new JSON;
    if(!$ENV{SSH_ORIGINAL_COMMAND}) {
        $j->pretty(1);
    }
    my $json = $j->encode($o);
    print $json;
}

sub progress {
    my $self = shift;
    my ($msg) = @_;
    print STDERR "SSH-API-PROGRESS:$msg\n";
    return;
}

sub caller_id {
    my $self = shift;
    return $ENV{SSHAPI_CALLER} // 'SSHAPI/UNKNOWN';
}

1;
